@@ -1,7 +1,11 @@
+import json
+import time
 import unittest
 
 from mock import Mock, patch
-from patroni.dcs.kubernetes import Kubernetes, KubernetesError, k8s_client, k8s_watch, RetryFailedError
+from patroni.dcs.kubernetes import Kubernetes, KubernetesError, k8s_client, RetryFailedError
+from threading import Thread
+from . import SleepException
 
 
 def mock_list_namespaced_config_map(self, *args, **kwargs):
@@ -15,32 +19,53 @@ def mock_list_namespaced_config_map(self, *args, **kwargs):
     metadata.update({'name': 'test-sync', 'annotations': {'leader': 'p-0'}})
     items.append(k8s_client.V1ConfigMap(metadata=k8s_client.V1ObjectMeta(**metadata)))
     metadata = k8s_client.V1ObjectMeta(resource_version='1')
-    return k8s_client.V1ConfigMapList(metadata=metadata, items=items)
+    return k8s_client.V1ConfigMapList(metadata=metadata, items=items, kind='ConfigMapList')
 
 
 def mock_list_namespaced_pod(self, *args, **kwargs):
     metadata = k8s_client.V1ObjectMeta(resource_version='1', name='p-0', annotations={'status': '{}'})
     items = [k8s_client.V1Pod(metadata=metadata)]
-    return k8s_client.V1PodList(items=items)
+    return k8s_client.V1PodList(items=items, kind='PodList')
 
 
-@patch.object(k8s_client.CoreV1Api, 'patch_namespaced_config_map', Mock())
-@patch.object(k8s_client.CoreV1Api, 'create_namespaced_config_map', Mock())
+def mock_config_map(*args, **kwargs):
+    mock = Mock()
+    mock.metadata.resource_version = '2'
+    return mock
+
+
+@patch.object(k8s_client.CoreV1Api, 'patch_namespaced_config_map', mock_config_map)
+@patch.object(k8s_client.CoreV1Api, 'create_namespaced_config_map', mock_config_map)
+@patch('kubernetes.client.api_client.ThreadPool', Mock(), create=True)
+@patch.object(Thread, 'start', Mock())
 class TestKubernetes(unittest.TestCase):
 
     @patch('kubernetes.config.load_kube_config', Mock())
     @patch.object(k8s_client.CoreV1Api, 'list_namespaced_config_map', mock_list_namespaced_config_map)
     @patch.object(k8s_client.CoreV1Api, 'list_namespaced_pod', mock_list_namespaced_pod)
+    @patch('kubernetes.client.api_client.ThreadPool', Mock(), create=True)
+    @patch.object(Thread, 'start', Mock())
     def setUp(self):
         self.k = Kubernetes({'ttl': 30, 'scope': 'test', 'name': 'p-0', 'retry_timeout': 10, 'labels': {'f': 'b'}})
-        with patch('time.time', Mock(return_value=1)):
+        self.assertRaises(AttributeError, self.k._pods._build_cache)
+        self.k._pods._is_ready = True
+        self.assertRaises(AttributeError, self.k._kinds._build_cache)
+        self.k._kinds._is_ready = True
+        self.k.get_cluster()
+
+    @patch('time.time', Mock(side_effect=[1, 10.9, 100]))
+    def test__wait_caches(self):
+        self.k._pods._is_ready = False
+        with self.k._condition:
+            self.assertRaises(RetryFailedError, self.k._wait_caches)
+
+    def test_get_cluster(self):
+        with patch.object(k8s_client.CoreV1Api, 'list_namespaced_config_map', mock_list_namespaced_config_map), \
+                patch.object(k8s_client.CoreV1Api, 'list_namespaced_pod', mock_list_namespaced_pod), \
+                patch('time.time', Mock(return_value=time.time() + 31)):
             self.k.get_cluster()
 
-    @patch.object(k8s_client.CoreV1Api, 'list_namespaced_config_map', mock_list_namespaced_config_map)
-    @patch.object(k8s_client.CoreV1Api, 'list_namespaced_pod', mock_list_namespaced_pod)
-    def test_get_cluster(self):
-        self.k.get_cluster()
-        with patch.object(k8s_client.CoreV1Api, 'list_namespaced_pod', Mock(side_effect=Exception)):
+        with patch.object(Kubernetes, '_wait_caches', Mock(side_effect=Exception)):
             self.assertRaises(KubernetesError, self.k.get_cluster)
 
     @patch('kubernetes.config.load_kube_config', Mock())
@@ -103,17 +128,12 @@ class TestKubernetes(unittest.TestCase):
         self.k.set_ttl(10)
         self.k.watch(None, 0)
         self.k.watch(None, 0)
-        with patch.object(k8s_watch.Watch, 'stream',
-                          Mock(side_effect=[Exception, [], KeyboardInterrupt,
-                                            [{'raw_object': {'metadata': {'resourceVersion': '2'}}}]])):
-            self.assertFalse(self.k.watch('1', 2))
-            self.assertRaises(KeyboardInterrupt, self.k.watch, '1', 2)
-            self.assertTrue(self.k.watch('1', 2))
 
     def test_set_history_value(self):
         self.k.set_history_value('{}')
 
     @patch('kubernetes.config.load_kube_config', Mock())
+    @patch('patroni.dcs.kubernetes.ObjectCache', Mock())
     @patch.object(k8s_client.CoreV1Api, 'patch_namespaced_pod', Mock(return_value=True))
     @patch.object(k8s_client.CoreV1Api, 'create_namespaced_endpoints', Mock())
     @patch.object(k8s_client.CoreV1Api, 'create_namespaced_service',
@@ -124,3 +144,31 @@ class TestKubernetes(unittest.TestCase):
         self.assertIsNotNone(k.patch_or_create_config({'foo': 'bar'}))
         self.assertIsNotNone(k.patch_or_create_config({'foo': 'bar'}))
         k.touch_member({'state': 'running', 'role': 'replica'})
+
+
+class TestCacheBuilder(unittest.TestCase):
+
+    @patch('kubernetes.config.load_kube_config', Mock())
+    @patch('kubernetes.client.api_client.ThreadPool', Mock(), create=True)
+    @patch.object(Thread, 'start', Mock())
+    def setUp(self):
+        self.k = Kubernetes({'ttl': 30, 'scope': 'test', 'name': 'p-0', 'retry_timeout': 10, 'labels': {'f': 'b'}})
+
+    @patch.object(k8s_client.CoreV1Api, 'list_namespaced_config_map', mock_list_namespaced_config_map)
+    @patch('patroni.dcs.kubernetes.ObjectCache._watch')
+    def test__build_cache(self, mock_response):
+        mock_response.return_value.read_chunked.return_value = [json.dumps(
+            {'type': 'MODIFIED', 'object': {'metadata': {
+                'name': self.k.config_path, 'resourceVersion': '2', 'annotations': {self.k._CONFIG: 'foo'}}}}
+        ).encode('utf-8'), ('\n' + json.dumps(
+            {'type': 'DELETED', 'object': {'metadata': {
+                'name': self.k.config_path, 'resourceVersion': '3'}}}
+        ) + '\n' + json.dumps(
+            {'type': 'MDIFIED', 'object': {'metadata': {'name': self.k.config_path}}}
+        ) + '\n' + json.dumps({'object': {'code': 410}}) + '\n').encode('utf-8')]
+        self.k._kinds._build_cache()
+
+    @patch('patroni.dcs.kubernetes.logger.error', Mock(side_effect=SleepException))
+    @patch('patroni.dcs.kubernetes.ObjectCache._build_cache', Mock(side_effect=Exception))
+    def test_run(self):
+        self.assertRaises(SleepException, self.k._pods.run)
